#include <Arduino_FreeRTOS.h>
#include <semphr.h>
#include <UTFTGLUE.h>
#include "FreeArial_round_16x24.h"
UTFTGLUE myGLCD(0, A2, A1, A3, A4, A0);

#define switch1 A5
#define switch2 A6 
#define switch3 A7
#define ledBarber1 26
#define ledBarber2 27
#define ledBarber3 28
#define RCK 10   // latchPin
#define SRCK 13  // clockPin
#define SIN 11   // dataPin
#define enterence 22
#define availableQ 24
#define unavailableQ 25
#define buzzer 12
#define timerMinBarber 2500
#define timerMaxBarber 7500

// Define the variables
byte queue = 0;
long timerBarber1, timerBarber2, timerBarber3, startBarber1, startBarber2, startBarber3;
String name[16] = { " Mihai  ", " Stefan ", "Cristian", " Andrei ",  " Costin ", 
                    " Horatiu", " Gabi   ", " Orlando", " Flavius", " Calin  ",
                    " Serban ", "  Radu  ", 
                    "  Paul  ", " Eduard ", " Codrut ", " Mircea "};
String queueDisplayed[8] = {"        ", "        ", "        ", "        ", 
                            "        ", "        ", "        ", "        "};
String clientBarber1, clientBarber2, clientBarber3;
int nameNewClient, insertPosition = 0;

// Tasks definition 
void clientsTask(void *pvParameters);
void barberTask1(void *pvParameters);
void barberTask2(void *pvParameters);
void barberTask3(void *pvParameters);
void queueTask(void *pvParameters);
void displayTask(void *pvParameters);

// Semaphores definition
SemaphoreHandle_t queueMutex;
SemaphoreHandle_t buzzerMutex;

void setup() {
  // Initialize the Serial port
  Serial.begin(9600);
  // Initialize the inputs and outputs
  pinMode(switch1, INPUT);
  pinMode(switch2, INPUT);
  pinMode(switch3, INPUT);
  pinMode(enterence, INPUT);
  pinMode(availableQ, OUTPUT);
  pinMode(unavailableQ, OUTPUT);
  pinMode(buzzer, OUTPUT);
  pinMode(RCK, OUTPUT);
  pinMode(SIN, OUTPUT);
  pinMode(SRCK, OUTPUT);
  pinMode(ledBarber1, OUTPUT);
  pinMode(ledBarber2, OUTPUT);
  pinMode(ledBarber3, OUTPUT);

  // Setup the LCD
  myGLCD.InitLCD();
  myGLCD.setFont(&FreeArial_round_16x24);
  myGLCD.clrScr();

  // Initialize the mutexes
  queueMutex = xSemaphoreCreateMutex();
  buzzerMutex = xSemaphoreCreateMutex();

  // Create tasks
  xTaskCreate(barberTask1, "barberTask1", 128, NULL, 1, NULL);
  xTaskCreate(barberTask2, "barberTask2", 128, NULL, 1, NULL);
  xTaskCreate(barberTask3, "barberTask3", 128, NULL, 1, NULL);
  xTaskCreate(clientsTask, "clientsTask", 128, NULL, 1, NULL);
  xTaskCreate(queueTask, "queueTask", 128, NULL, 1, NULL);
  xTaskCreate(displayTask, "displayTask", 4096, NULL, 1, NULL);
  // Start tasks schedule
  vTaskStartScheduler(); 
}

void loop() {}

void emptyingQueue(int barber) {
  String copy = queueDisplayed[0]; // Copy the first element from queue
  for (int i = 0; i < 7; i++) { // Delete the first element 
    queueDisplayed[i] = queueDisplayed[i + 1];
  }
  queueDisplayed[7] = "        ";
  insertPosition--;
  // Set the name to display for the barber who called the function
  switch(barber) {
    case 1: {
      clientBarber1 = copy;
      break;
    }
    case 2: {
      clientBarber2 = copy;
      break;
    }
    case 3: {
      clientBarber3 = copy;
      break;
    }
  }
}

void barberTask1(void *pvParameters) {
  while (1) {
    if (queue && digitalRead(switch1) && 
       (xSemaphoreTake(queueMutex, portMAX_DELAY) == pdTRUE)) { 
      // Condition if queue is not empty, the barber switch is in the position 
      // on and mutex it has not been taken
      queue = queue << 1; // Take client from queue
      xSemaphoreGive(queueMutex); // Release the mutex
      timerBarber1 = random(timerMinBarber, timerMaxBarber); 
      // Choose a random time for barber services
      startBarber1 = millis() + timerBarber1; 
      // Variabile use to display the remaining time
      digitalWrite(ledBarber1, HIGH); // Turn on the LED Barber
      emptyingQueue(1); // Call funtion to take a client from queue 
      vTaskDelay(timerBarber1 / portTICK_PERIOD_MS); 
      // Time for barber services 
      digitalWrite(ledBarber1, LOW); // Turn off the LED
      if (xSemaphoreTake(buzzerMutex, portMAX_DELAY) == pdTRUE) { 
        // Check if the buzzer is not in use, otherwise, use them
        tone(buzzer, 2150, 250);
        xSemaphoreGive(buzzerMutex);
      }
    } else {
      xSemaphoreGive(queueMutex); // Release queue mutex
    }
    vTaskDelay(50 / portTICK_PERIOD_MS);
  }
}

void barberTask2(void *pvParameters) {
  // The same functionality like barberTask1
  while (1) {
    if (queue && digitalRead(switch2) && 
       (xSemaphoreTake(queueMutex, portMAX_DELAY) == pdTRUE)) {
      queue = queue << 1;
      xSemaphoreGive(queueMutex);
      timerBarber2 = random(timerMinBarber, timerMaxBarber);
      startBarber2 = millis() + timerBarber2;
      digitalWrite(ledBarber2, HIGH);
      emptyingQueue(2);
      vTaskDelay(timerBarber2 / portTICK_PERIOD_MS);
      digitalWrite(ledBarber2, LOW);
      if (xSemaphoreTake(buzzerMutex, portMAX_DELAY) == pdTRUE) {
        tone(buzzer, 2150, 250);
        xSemaphoreGive(buzzerMutex);
      }
    } else {
      xSemaphoreGive(queueMutex);
    }
    vTaskDelay(50 / portTICK_PERIOD_MS);
  }
}

void barberTask3(void *pvParameters) {
  // The same functionality like barberTask1
  while (1) {
    if (queue && digitalRead(switch3) && 
       (xSemaphoreTake(queueMutex, portMAX_DELAY) == pdTRUE)) {
      queue = queue << 1;
      xSemaphoreGive(queueMutex);
      timerBarber3 = random(timerMinBarber, timerMaxBarber);
      startBarber3 = millis() + timerBarber3;
      digitalWrite(ledBarber3, HIGH);
      emptyingQueue(3);
      vTaskDelay(timerBarber3 / portTICK_PERIOD_MS);
      digitalWrite(ledBarber3, LOW);
      if (xSemaphoreTake(buzzerMutex, portMAX_DELAY) == pdTRUE) {
        tone(buzzer, 2150, 250);
        xSemaphoreGive(buzzerMutex);
      }
    } else {
      xSemaphoreGive(queueMutex);
    }
    vTaskDelay(50 / portTICK_PERIOD_MS);
  }
}

void queueTask(void *pvParameters) {
  // Task used for control queue LEDs 
  while (1) { 
    digitalWrite(RCK, LOW);
    shiftOut(SIN, SRCK, MSBFIRST, queue);
    digitalWrite(RCK, HIGH);
    vTaskDelay(50 / portTICK_PERIOD_MS);
  }
}

void displayTask(void *pvParameters) {
  while (1) {
    int minute, seconds;
    myGLCD.setColor(255, 255, 255); // Set white color for the lines
    myGLCD.drawLine(155, 5, 155, 115); 
    myGLCD.drawLine(320, 5, 320, 115);
    myGLCD.drawLine(5, 115, 475, 115);
    myGLCD.drawLine(240, 115, 240, 315);
    myGLCD.setBackColor(0, 0, 0); // Set the backgraound color

    if(digitalRead(switch1) && digitalRead(ledBarber1) || 
      !digitalRead(switch1) && digitalRead(ledBarber1)) {
      myGLCD.setColor(255, 255, 255); // Set white color for the barber client
      myGLCD.print(clientBarber1, 5, 43);
      // Calculate the time for current client 
      seconds = startBarber1 - millis();
      minute = startBarber1 - millis();
      if (seconds > 0 && minute > 0) {
        myGLCD.print("  0" + String(minute/1000) + ":" +
                     String(seconds/10%100) + "  ", 5, 83);
      }
      else {
        myGLCD.print("  00:00  ", 5, 83);
      }
      myGLCD.setColor(245, 165, 73); // Set orange color Barber1 text
    }
    else if(digitalRead(switch1) && !digitalRead(ledBarber1)) {
      myGLCD.setColor(255, 255, 255); // Set white color Liber text
      myGLCD.print("  Liber  ", 5, 43);
      myGLCD.print("  00:00  ", 5, 83);
      myGLCD.setColor(0, 255, 0); // Set green color Barber1 text
    }
    else {
      myGLCD.setColor(255, 255, 255); // Set white color Absent text
      myGLCD.print("  Absent ", 0, 43);
      myGLCD.print("        ", 5, 83);
      myGLCD.setColor(255, 0, 0); // Set Red color Barber1 text
      
    }
    myGLCD.print("Barbier 1", 5, 8);

    // Identic with Barber1
    if(digitalRead(switch2) && digitalRead(ledBarber2) || 
      !digitalRead(switch2) && digitalRead(ledBarber2)) {
      myGLCD.setColor(255, 255, 255);
      myGLCD.print(clientBarber2, 165, 43);
      seconds = startBarber2 - millis();
      minute = startBarber2 - millis();
      if (seconds > 0 && minute > 0) {
        myGLCD.print("  0" + String(minute/1000) + ":" + 
                     String(seconds/10%100) + "  ", 165, 83);
      }
      else {
        myGLCD.print("  00:00  ", 165, 83);
      }
      myGLCD.setColor(245, 165, 73);
    }
    else if(digitalRead(switch2) && !digitalRead(ledBarber2)) {
      myGLCD.setColor(255, 255, 255);
      myGLCD.print("  Liber  ", 165, 43);
      myGLCD.print("  00:00  ", 165, 83);
      myGLCD.setColor(0, 255, 0);
    }
    else {
      myGLCD.setColor(255, 255, 255);
      myGLCD.print("  Absent ", 165, 43);
      myGLCD.print("        ", 165, 83);
      myGLCD.setColor(255, 0, 0);
    }
    myGLCD.print("Barbier 2", 160, 8);

    // Identic with Barber1
    if(digitalRead(switch3) && digitalRead(ledBarber3) || 
      !digitalRead(switch3) && digitalRead(ledBarber3)) {
      myGLCD.setColor(255, 255, 255);
      myGLCD.print(clientBarber3, 330, 43);
      seconds = startBarber3 - millis();
      minute = startBarber3 - millis();
      if (minute > 0 && seconds > 0) {
        myGLCD.print("  0" + String(minute/1000) + ":" + 
                     String(seconds/10%100) + "  ", 330, 83);
      }
      else {
        myGLCD.print("  00:00  ", 330, 83);
      }
      myGLCD.setColor(245, 165, 73);
    }
    else if(digitalRead(switch3) && !digitalRead(ledBarber3)) {
      myGLCD.setColor(255, 255, 255);
      myGLCD.print("  Liber  ", 330, 43);
      myGLCD.print("  00:00  ", 330, 83);
      myGLCD.setColor(0, 255, 0);
    }
    else {
      myGLCD.setColor(255, 255, 255);
      myGLCD.print("  Absent ", 325, 43);
      myGLCD.print("        ", 330, 83);
      myGLCD.setColor(255, 0, 0);
    }
    myGLCD.print("Barbier 3", 330, 8);

    myGLCD.setColor(255, 255, 255); 
    // Set color to white for clients name queue
    for(int i = 0; i < 8; i++) {
      String client = String(i + 1);
      client += ". ";
      myGLCD.print(client, 5 + (i / 4)*250, 135 + (i % 4)*50);
      myGLCD.print(queueDisplayed[i], 40 + (i / 4)*250, 135 + (i % 4)*50);
    }

    vTaskDelay(10 / portTICK_PERIOD_MS);
  }
}

void clientsTask(void *pvParameters) {
  while (1) {
    if (digitalRead(enterence) && 
        xSemaphoreTake(queueMutex, portMAX_DELAY) == pdTRUE) { 
      // Verify if a new client has entered and check if the queue mutex 
      // has been taken
      if (queue == 255 && xSemaphoreTake(buzzerMutex, portMAX_DELAY) == pdTRUE) { 
        // If the queue is full, and the buzzer is not in use, turn on the 
        // red LED and activate the buzzer
        tone(buzzer, 2800, 200);
        xSemaphoreGive(buzzerMutex);
        digitalWrite(unavailableQ, HIGH);
        vTaskDelay(300 / portTICK_PERIOD_MS);
        digitalWrite(unavailableQ, LOW);
        xSemaphoreGive(queueMutex);
      } else {// Otherwise, add a new client in queue and update the LEDs queue
        byte shift = queue >> 1;
        queue = 128 + shift;
        xSemaphoreGive(queueMutex);
        nameNewClient = random(0, 15); 
        // Choose a random name from the name vector defined
        queueDisplayed[insertPosition] = name[nameNewClient]; 
        // Add the newest client in displayedQueue
        insertPosition++;
        if (xSemaphoreTake(buzzerMutex, portMAX_DELAY) == pdTRUE) { 
          // Activate the buzzer if is not in use
          digitalWrite(availableQ, HIGH);
          vTaskDelay(300 / portTICK_PERIOD_MS);
          digitalWrite(availableQ, LOW);
          tone(buzzer, 1500, 50);
          vTaskDelay(200 / portTICK_PERIOD_MS);
          tone(buzzer, 2500, 50);
          xSemaphoreGive(buzzerMutex);
        }
      }
    }
    else{
      xSemaphoreGive(queueMutex);
    }
    vTaskDelay(50 / portTICK_PERIOD_MS);
  }
}

